/*
 * can_transfer.c
 *
 * Created: 12.02.2012 11:28:53
 *  Author: Privat
 */ 

#include "can_transfer.h"
#include <avr/wdt.h>
#include <string.h>

CAN_BUFFER	can_buffer;
uint16_t	active_msgid = -1;
uint16_t	active_command = 0;
uint8_t		message_length = 0;

uint8_t GetNextMsgFromBuffer(can_t *msg, uint8_t a_msgid, uint8_t a_command)
{
	uint8_t i;
	uint8_t found = false;
	
	// check if a message is in the buffer
	if (can_buffer.Count > 0)
	{		
		// set index to start
		i = 0;
		// find next message for the active msg id
		// return only messages what are matching to the msg id and command
		// if a transfer is active
		if (a_msgid != -1)
		{			
			for (i = 0; i < CAN_BUFFER_SIZE; i++)
			{
				if ((can_buffer.canmsg[i].id == a_msgid) && ((can_buffer.canmsg[i].data[1] & COMMAND_MASK) == a_command))
				{
					found = true;
					break;		
				}				
			}
			if (!found)
				return false;				
		} 

		// copy message if a_msgid == -1 or (a_msgid != -1 and found == true)
		memcpy(msg, &can_buffer.canmsg[i], sizeof(can_t));
		// decrement counter
		can_buffer.Count--;	
		
		// if more messages are in the buffer
		// move them forward to the last pickup position
		while(i < can_buffer.Count)
		{
			memcpy(&can_buffer.canmsg[i], &can_buffer.canmsg[i+1], sizeof(can_t));
			i++;	
		}
		
		return true;
	}
	
	return false;
}

uint8_t BufferCANMsg(can_t *msg)
{
	// check if buffer is full
	if (can_buffer.Count >= CAN_BUFFER_SIZE)
	{
		return false;
	} 
	else
	{
		// buffer message for later handling					
		memcpy(&can_buffer.canmsg[can_buffer.Count], msg, sizeof(can_t));
		can_buffer.Count++;	
		return true;		
	}
}

uint8_t CheckForNewCANMessage(can_t *msg)
{
	if (can_check_message())
	{
		if (can_get_message(msg))
			return true;
	}
	return false;
}

uint8_t MessageHandling(CAN_HEADER *can_transfer_p, can_t *msg, uint8_t *Message_Counter, uint8_t data[], uint8_t *len)
{
	// check if active transfer is running
	if (active_msgid == -1)		
	{
		// new transfer starting
		// set msg id and command
		active_msgid = msg->id;
		active_command = msg->data[1] & COMMAND_MASK;
		message_length = 0;
	}
	else
	// check if received msg id and if it match
	// chekc if the command is matching too
	if (active_msgid == msg->id)
	{
		// check if command is matching
		if (active_command != (msg->data[1] & COMMAND_MASK))
		{
			// buffer message for later handling					
			BufferCANMsg(msg);
			return false;		
		}			
	}
	// else msg id does not match to running
	// transfer, buffer it for later handling
	else
	{
		// buffer message for later handling					
		BufferCANMsg(msg);
		return false;		
	}
	
	/*
	(*Message_Counter)++;
	if (*Message_Counter != canmsg.data[2])
	{
		// received message with wrong
		// message number, return expected
		// number
		can_transfer.Board_Identifier = CAN_NODE_NR;
		can_transfer.CAN_MessageID = active_msgid;
		can_transfer.T_OR_C = active_command | WRONG_NUMBER_REPSONSE;
		SendCANMessage(&can_transfer, Message_Counter, NULL, 0);
		return false;
	}
	*/
	if (Message_Counter != NULL)
		*Message_Counter = msg->data[2];	
		
	// check if new block starts
	if (msg->data[3] & START_OF_MESSAGE_MASK)
		message_length = 0;
		
	// check if message includes data bytes
	if (((msg->length - CAN_HEADER_SIZE) > 0) && (data != NULL))
		memcpy(&data[message_length], &msg->data[CAN_HEADER_SIZE], msg->length - CAN_HEADER_SIZE);
			
	message_length += msg->length - CAN_HEADER_SIZE;
		
	// check if this was the last message
	if ((msg->data[3] & ~START_OF_MESSAGE_MASK) == 0)
	{
		// return needed message information
		can_transfer_p->CAN_MessageID = active_msgid;
		can_transfer_p->Board_Identifier = msg->data[0];		
		can_transfer_p->T_OR_C = msg->data[1];
			
		if (len != NULL)		
			*len = message_length;
				
		// reset active transfer to be able to receive next
		// transfer
		active_msgid = -1;
		active_command = 0;
		message_length = 0;
		return true;
	}	
	return false;		
}

void InitCANBus(uint8_t bitrate, const prog_uint8_t *filter)
{
	
	// init can buffer
	memset(&can_buffer, 0x00, sizeof(CAN_BUFFER));
		
	can_init(bitrate);											// Initialize MCP2515
	can_static_filter(filter);									// Load filters and masks	
}

uint8_t CANTranserActive(void)
{
	if (active_msgid == -1)
	{
		return false;
	} 
	else
	{
		return true;
	}
}

uint8_t WaitforCANResponse(CAN_HEADER *can_transfer_p, uint8_t *Message_Counter, uint8_t data[], uint8_t *len)
{
	can_t canmsg;
	uint8_t timeout = 100;	
	
	// finally wait for a response:
	while (1)			
	{
		if (CheckForNewCANMessage(&canmsg))
		{
			if (canmsg.data[0] == can_transfer_p->Board_Identifier)
			{
				MessageHandling(can_transfer_p, &canmsg, Message_Counter, data, len);
				return (can_transfer_p->T_OR_C & ~COMMAND_MASK);
			}
			else
			{
				// buffer message for later handling					
				BufferCANMsg(&canmsg);							
			}				
		}		

		_delay_ms(5);
		timeout--;
		
		if (timeout < 1)
			return ERROR_RESPONSE;
	}		
}

uint8_t ReceiveCANMessage(CAN_HEADER *can_transfer_p, uint8_t *Message_Counter, uint8_t data[], uint8_t *len)
{
	can_t canmsg;
	uint8_t newcanmsg = false;
	
	// first check can message buffer
	if (GetNextMsgFromBuffer(&canmsg, active_msgid, active_command))
	{
		newcanmsg = true;
	} 
	else
	if (CheckForNewCANMessage(&canmsg))
	{
		newcanmsg = true;
	}

	// message handling
	if (newcanmsg)
		return MessageHandling(can_transfer_p, &canmsg, Message_Counter, data, len);
	
	return false;
}

uint8_t SendCANMessage(CAN_HEADER *can_transfer_p, uint8_t *Message_Counter, uint8_t data[], uint8_t len)
{
	can_t canmsg, canmsgresponse;
	uint8_t data_len;
	uint8_t last_len;
	uint8_t needed_msg_count;
	
	// get number of needed messages
	if ( len > CAN_DATA_SIZE)
	{
		needed_msg_count = len / CAN_DATA_SIZE;
		
		if ((len % CAN_DATA_SIZE) != 0)
			needed_msg_count++;			
	} else { needed_msg_count = 1; }
		
	// check if len > 4 * 0x0F
	// message is too long!!
	if (needed_msg_count > 0x0F)
		return ERROR_RESPONSE;
		
	// set basics in can message:
	canmsg.flags.rtr = 0;
	canmsg.id = can_transfer_p->CAN_MessageID | CAN_NODE_NR;
	canmsg.length = CAN_HEADER_SIZE;
	canmsg.data[0] = can_transfer_p->Board_Identifier;
	canmsg.data[1] = can_transfer_p->T_OR_C;	
	canmsg.data[3] = START_OF_MESSAGE_MASK;
	
	if (data == NULL)
		len = 0;

	data_len = 0;

	// send messages while needed_msg_count > 0
	while (needed_msg_count > 0)
	{
		needed_msg_count--;
		
		if (Message_Counter != NULL)
		{	
			(*Message_Counter)++;			
			canmsg.data[2] = *Message_Counter;
		}
				
		canmsg.data[3] &= 0xF0;
		canmsg.data[3] |= needed_msg_count;
		
		// copy data
		while ((data_len < len) && (canmsg.length < (CAN_DATA_SIZE + CAN_HEADER_SIZE)))
		{
			canmsg.data[canmsg.length] = data[data_len];
			data_len++;
			canmsg.length++;
		}
		
		last_len = canmsg.length - CAN_HEADER_SIZE;		
		
		while(!can_send_message(&canmsg))
			wdt_reset();
			
		// clear SoB for next message
		canmsg.data[3] &= ~START_OF_MESSAGE_MASK;
		// reset canmsg length for next message
		canmsg.length = CAN_HEADER_SIZE;
		
		// check for WRONG_NUMBER_REPSONSE or ERROR_RESPONSE
		while (CheckForNewCANMessage(&canmsgresponse))
		{	
			if (((canmsgresponse.data[1] & COMMAND_MASK) == (can_transfer_p->T_OR_C & COMMAND_MASK)) && 
				((canmsgresponse.data[1] & ~COMMAND_MASK) == WRONG_NUMBER_REPSONSE))
			{
				// can not handle missing messages if message counter is not used
				if (Message_Counter != NULL)
					continue;
					
				// wrong message number
				// go back to the requested message number
				while (*Message_Counter > canmsgresponse.data[2])
				{
					(*Message_Counter)--;
					needed_msg_count++;
					// check if active transfer includes data
					if (len > 0)
					{
						// sub last message size from data_len
						// can be 1 .. 4 bytes
						if (last_len > 0)
						{
							data_len -= last_len;
							last_len = 0;
						} 
						// sub last - 1 message len from data_len
						// have to be 4 bytes
						else
						{
							data_len -= CAN_DATA_SIZE;
						}
					}					
				}
			} 
			else
			if (((canmsgresponse.data[1] & COMMAND_MASK) == (can_transfer_p->T_OR_C & COMMAND_MASK)) && 
				((canmsgresponse.data[1] & ~COMMAND_MASK) == ERROR_RESPONSE))
			{
				// error response
				return ERROR_RESPONSE;
			} 
			else								
			{
				// buffer message for later handling					
				BufferCANMsg(&canmsgresponse);
			}						
		}			
	}	
	return SUCCESSFULL_RESPONSE;
}
